from inspect import signature


def style_function(name, value):
    nargs = len(signature(value).parameters)

    def callback(node, data, i, group):
        args = [data, i, group][:nargs]
        current_value = node.get("style", "")
        new_value = value(*args)
        node.set("style", f"{current_value}{name}:{new_value};")

    return callback


def style_constant(name, value):
    def callback(node, data, i, group):
        current_value = node.get("style", "")
        node.set("style", f"{current_value}{name}:{value};")

    return callback


def style_value(style, name):
    if style.endswith(";"):
        style = style[:-1]
    attrs = {
        property: value
        for property, value in (desc.split(":") for desc in style.split(";"))
    }
    return attrs.get(name)



from __future__ import annotations

import math
from typing import Any, TypeVar

from .linear import LinearBase

T = TypeVar("T")


class Identity(LinearBase):
    """
    Build a new identity scale with the specified range (and by extension, domain).
    """

    def __init__(self, domain=None):
        self._domain = list(domain) if domain is not None else [0, 1]
        self._unknown = None

    def __call__(self, x: T) -> T:
        """
        Returns same value if valid type

        Parameters
        ----------
        x : T
            Input value

        Returns
        -------
        T
            Input value if valid type
        """
        unvalid_type = x is None or (isinstance(x, float) and math.isnan(x))
        return self._unknown if unvalid_type else x

    def invert(self, x: T) -> T:
        """
        Returns same value if valid type

        Parameters
        ----------
        x : T
            Input value

        Returns
        -------
        T
            Input value if valid type
        """
        return self(x)

    def set_domain(self, domain: list[T]) -> Identity:
        """
        Sets the scale’s domain

        Parameters
        ----------
        domain : list[T]
            Domain

        Returns
        -------
        Identity
            Itself
        """
        self._domain = list(domain)
        return self

    @property
    def domain(self):
        return self._domain

    def set_range(self, range_vals: list[T]) -> Identity:
        """
        Sets the scale’s range

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Identity
            Itself
        """
        return self.set_domain(range_vals)

    @property
    def range(self):
        return self._domain

    def set_unknown(self, unknown: Any) -> Identity:
        """
        Sets the output value of the scale for undefined
        or NaN input values.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        Identity
            Itself
        """
        self._unknown = unknown
        return self

    @property
    def unknown(self):
        return self._unknown

    def copy(self):
        return Identity(self._domain).set_unknown(self._unknown)


def scale_identity(values: list[T] | None = None) -> Identity:
    """
    Build a new identity scale with the specified range
    (and by extension, domain).

    Parameters
    ----------
    values : list[T] | None
        Values to set

    Returns
    -------
    Identity
        Scale object
    """
    return Identity(domain=values)



import math


class LinearCurve:
    def __init__(self, context):
        self._context = context
        self._line = math.nan

    def area_start(self):
        self._line = 0

    def area_end(self):
        self._line = math.nan

    def line_start(self):
        self._point = 0

    def line_end(self):
        if not math.isnan(self._line) or (self._line != 0 and self._point == 1):
            self._context.close_path()
        self._line = 1 - self._line

    def point(self, x, y):
        if self._point == 0:
            self._point = 1
            if not math.isnan(self._line) and self._line:
                self._context.line_to(x, y)
            else:
                self._context.move_to(x, y)
        elif self._point == 1:
            self._point = 2
            self._context.line_to(x, y)
        else:
            self._context.line_to(x, y)



from __future__ import annotations

import math
from collections.abc import Callable
from typing import overload

from .continuous import Transformer, identity
from .init import init_range
from .linear import LinearBase
from .number import number


def sign(x):
    return -1 if x < 0 else 1


def square(x):
    return sign(x) * x * x


def unsquare(x):
    return sign(x) * math.sqrt(abs(x))


class ScaleRadial(Transformer, LinearBase):
    """
    Radial scales are a variant of linear scales where the range
    is internally squared so that an input value corresponds linearly
    to the squared output value. These scales are useful when you want
    the input value to correspond to the area of a graphical mark and
    the mark is specified by radius, as in a radial bar chart.
    Radial scales do not support interpolate.

    Parameters
    ----------
    t : Callable
        Tranform function
    u : Callable
        Untranform function
    """

    def __init__(self, t: Callable = identity, u: Callable = identity):
        super().__init__(t, u)
        self._range_vals = [0, 1]
        self._round = False
        self._unknown = None

    def __call__(self, x: int | float) -> int | float:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        x : int | float
            Input value

        Returns
        -------
        int | float
            Corresponding value from the range
        """
        y = unsquare(super().__call__(x))
        if isinstance(y, float) and math.isnan(y):
            return self._unknown
        elif self._round:
            return round(y)
        else:
            return y

    def invert(self, y: int | float) -> int | float:
        """
        Given a value from the range, returns the corresponding value
        from the domain. Inversion is useful for interaction, say to
        determine the data value corresponding to the position of the mouse.

        Parameters
        ----------
        y : int | float
            Input value

        Returns
        -------
        int | float
            Corresponding value from the domain
        """
        return super().invert(square(y))

    def set_range(self, range_vals: list[int | float]) -> ScaleRadial:
        """
        Sets the scale's range to the specified array of values

        Parameters
        ----------
        range_vals : list[int | float]
            Range values

        Returns
        -------
        ScaleRadial
            Itself
        """
        self._range_vals = [number(x) for x in range_vals]
        super().set_range([square(x) for x in self._range_vals])
        return self

    @property
    def range(self):
        return self._range_vals.copy()

    def set_range_round(self, range_vals: list[int | float]) -> ScaleRadial:
        """
        Sets the scale's range to the specified array of values
        and sets scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[int | float]
            Range values

        Returns
        -------
        ScaleRadial
            Itself
        """
        return super().set_range(range_vals).set_round(True)

    def set_round(self, round_val: bool) -> ScaleRadial:
        """
        Enables or disables rounding accordingly

        Parameters
        ----------
        round_val : bool
            Round value

        Returns
        -------
        ScaleRadial
            Itself
        """
        self._round = bool(round_val)
        return self

    @property
    def round(self):
        return self._round

    def copy(self):
        return (
            ScaleRadial()
            .set_domain(self.domain)
            .set_range(self.range)
            .set_round(self.round)
            .set_clamp(self.clamp)
            .set_unknown(self._unknown)
        )


@overload
def scale_radial() -> ScaleRadial: ...


@overload
def scale_radial(range_vals: list[int | float]) -> ScaleRadial: ...


@overload
def scale_radial(
    domain: list[int | float], range_vals: list[int | float]
) -> ScaleRadial: ...


def scale_radial(*args):
    """
    Builds  a new radial scale with the specified domain and range.

    Parameters
    ----------
    domain : list[int | float]
        Array of numbers
    range_vals : list[int | float]
        Array of values

    Returns
    -------
    ScaleRadial
        Scale object

    Examples
    --------

    >>> d3.scale_radial([100, 200], [0, 480])
    """
    scale = ScaleRadial()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from __future__ import annotations

from collections.abc import Callable
from inspect import signature
from typing import Literal, Type

from ..scale.continuous import Transformer
from ..scale.diverging import Diverging
from ..scale.sequential import Sequential
from ..selection.selection import Selection

TOP = 1
RIGHT = 2
BOTTOM = 3
LEFT = 4
EPSILON = 1e-6


def translate_x(x):
    return f"translate({x}, 0)"


def translate_y(y):
    return f"translate(0, {y})"


def number(scale):
    def f(d):
        return float(scale(d))

    return f


def center(scale, offset):
    offset = max(0, scale.bandwidth - offset * 2) / 2
    if scale.round:
        offset = round(offset)
    return lambda d: float(scale(d)) + offset


# def entering(context):
#     return not hasattr(context, "__axis")


class Axis:
    """
    Builds a new oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.

    Parameters
    ----------
    orient : Literal[1, 2, 3, 4]
        Orientation
    scale : Type[Transformer | Sequential | Diverging]
        Scaler
    """

    def __init__(
        self,
        orient: Literal[1, 2, 3, 4],
        scale: Type[Transformer | Sequential | Diverging],
    ):
        self._scale = scale
        self._orient = orient
        self._tick_arguments = []
        self._tick_values = None
        self._tick_format = None
        self._tick_size_inner = 6
        self._tick_size_outer = 6
        self._tick_padding = 3
        self._offset = 0.5
        self._k = -1 if orient in [TOP, LEFT] else 1
        self._x = "x" if orient in [LEFT, RIGHT] else "y"
        self._transform = translate_x if orient in [TOP, BOTTOM] else translate_y

    def __call__(self, context: Selection):
        """
        Render the axis to the given context, which may be either a selection
        of SVG containers (either SVG or G elements) or a corresponding transition.

        Parameters
        ----------
        context : Selection
            Context

        Examples
        --------

        >>> svg.append("g")
        ...     .attr("transform", f"translate(0, {height - margin.bottom})")
        ...     .call(d3.axis_bottom(x))

        Notes
        -----
        Transitions are not yet implemented
        """
        if self._tick_values is not None:
            values = self._tick_values
        elif hasattr(self._scale, "ticks"):
            nargs = len(signature(self._scale.ticks).parameters)
            args = self._tick_arguments[:nargs]
            values = self._scale.ticks(*args)
        else:
            values = self._scale.domain

        if self._tick_format is not None:
            format_func = self._tick_format
        elif hasattr(self._scale, "tick_format"):
            nargs = len(signature(self._scale.tick_format).parameters)
            args = self._tick_arguments[:nargs]
            format_func = self._scale.tick_format(*args)
        else:

            def format_func(d):
                return d

        spacing = max(self._tick_size_inner, 0) + self._tick_padding
        range_values = self._scale.range
        range0 = float(range_values[0]) + self._offset
        range1 = float(range_values[-1]) + self._offset

        if hasattr(self._scale, "bandwidth"):
            position = center(self._scale.copy(), self._offset)
        else:
            position = number(self._scale.copy())

        selection = context.selection() if hasattr(context, "selection") else context
        path = selection.select_all(".domain").data([None])
        tick = selection.select_all(".tick").data(values, self._scale).order()
        tick_exit = tick.exit()
        tick_enter = tick.enter().append("g").attr("class", "tick")
        line = tick.select("line")
        text = tick.select("text")

        path = path.merge(
            path.enter()
            .insert("path", ".tick")
            .attr("class", "domain")
            .attr("stroke", "currentColor")
        )

        tick = tick.merge(tick_enter)

        line = line.merge(
            tick_enter.append("line")
            .attr("stroke", "currentColor")
            .attr(f"{self._x}2", self._k * self._tick_size_inner)
        )

        text = text.merge(
            tick_enter.append("text")
            .attr("fill", "currentColor")
            .attr(self._x, self._k * spacing)
            .attr(
                "dy",
                "0em"
                if self._orient == TOP
                else "0.71em"
                if self._orient == BOTTOM
                else "0.32em",
            )
        )

        # TODO : transition implementation
        # if context != selection:
        #     path = path.transition(context)
        #     tick = tick.transition(context)
        #     line = line.transition(context)
        #     text = text.transition(context)
        #
        #     def transform(d):
        #         d = position(d)
        #         if d is not None and math.isfinite(d):
        #             return self._transform(position(d) + self._offset)
        #
        #     tick_exit = (
        #         tick_exit.transition(context)
        #                  .attr("opacity", EPSILON)
        #                  .attr("transform", transform)
        #     )
        #
        #     def transform(d): # TODO
        #         return None
        #
        #     tick_enter.attr("opacity", EPSILON).attr("transform", transform)

        tick_exit.remove()

        if self._orient == LEFT or self._orient == RIGHT:
            if self._tick_size_outer:
                path.attr(
                    "d",
                    f"M{self._k * self._tick_size_outer},{range0}H{self._offset}V{range1}H{self._k * self._tick_size_outer}",
                )
            else:
                path.attr("d", f"M{self._offset},{range0}V{range1}")
        else:
            if self._tick_size_outer:
                path.attr(
                    "d",
                    f"M{range0},{self._k * self._tick_size_outer}V{self._offset}H{range1}V{self._k * self._tick_size_outer}",
                )
            else:
                path.attr("d", f"M{range0},{self._offset}H{range1}")

        def transform(d, *args):
            return self._transform(position(d) + self._offset)

        tick.attr("opacity", 1).attr("transform", transform)

        line.attr(f"{self._x}2", self._k * self._tick_size_inner)

        text.attr(self._x, self._k * spacing).text(format_func)

        (
            selection.attr("fill", "none")  # .filter(entering(context))
            .attr("font-size", 10)
            .attr("font-family", "sans-serif")
            .attr(
                "text-anchor",
                "start"
                if self._orient == RIGHT
                else "end"
                if self._orient == LEFT
                else "middle",
            )
        )

    def set_scale(self, scale: Type[Transformer | Sequential | Diverging]) -> Axis:
        """
        Sets scale value

        Parameters
        ----------
        scale : Type[Transformer | Sequential | Diverging]
            New scale

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> x_axis = d3.axis_bottom().scale(x)
        """
        self._scale = scale
        return self

    @property
    def scale(self) -> Type[Transformer | Sequential | Diverging]:
        return self._scale

    def set_ticks(self, *ticks: int | str | Callable) -> Axis:
        """
        Tick values will be passed to :code:`scale.ticks` and
        :code:`scale.tick_format` when :code:`Axis.__call__` is called

        Parameters
        ----------
        *ticks : int | str | Callable
            It depends on :code:`scale` type. Most of the time, it is the
            *count* for the number of ticks and optional format specifier

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        To generate twenty ticks with SI-prefix formatting on a linear scale :

        >>> axis.set_ticks(20, "s")

        To generate ticks every fifteen minutes with a time scale :

        >>> axis.set_ticks(d3.time_minute.every(15))

        Note the similarities with :code:`set_tick_arguments` :

        >>> axis.set_ticks(10) # same as axis.set_tick_arguments([10])
        """
        self._tick_arguments = list(ticks)
        return self

    def set_tick_arguments(self, tick_arguments: list[int | str | Callable]) -> Axis:
        """
        Tick arguments will be passed to :code:`scale.ticks` and
        :code:`scale.tick_format` when :code:`Axis.__call__` is called

        Parameters
        ----------
        tick_arguments : list[int | str | Callable]
            It depends on :code:`scale` type. Most of the time, it is the
            *count* for the number of ticks and optional format specifier

        Returns
        -------
            Itself

        Examples
        --------

        To generate twenty ticks with SI-prefix formatting on a linear scale :

        >>> axis.set_tick_arguments([20, "s"])

        To generate ticks every fifteen minutes with a time scale :

        >>> axis.set_tick_arguments([d3.time_minute.every(15)])
        """
        self._tick_arguments = list(tick_arguments)
        return self

    @property
    def tick_arguments(self) -> list[int | str | Callable]:
        return self._tick_arguments.copy()

    def set_tick_values(self, tick_values: list[int | float]) -> Axis:
        """
        Ticks values are used for ticks rather than the scale’s
        automatic tick generator.

        Parameters
        ----------
        tick_values : list[int | float]
            Tick values

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> axis = d3.axis_bottom(x).set_tick_values([1, 2, 3, 5, 8, 13, 21])
        """
        self._tick_values = tick_values
        return self

    @property
    def tick_values(self) -> list[int | float]:
        return self._tick_values

    def set_tick_format(self, tick_format: Callable) -> Axis:
        """
        Sets the tick format function and returns the axis.

        Parameters
        ----------
        tick_format : Callable
            Tick formatter

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        For example, to display integers with comma-grouping for thousands :

        >>> axis.set_tick_format(d3.format(",.0f"))
        """
        self._tick_format = tick_format
        return self

    @property
    def tick_format(self) -> Callable:
        return self._tick_format

    def set_tick_size(self, tick_size: int) -> Axis:
        """
        Sets the inner and outer tick size to the specified
        value and returns the axis.

        Parameters
        ----------
        tick_size : int
            Tick size

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> d3.axis_bottom().set_tick_size(0)
        """
        self._tick_size_inner = self._tick_size_outer = tick_size
        return self

    def set_tick_size_inner(self, tick_size_inner: int) -> Axis:
        """
        Sets the inner tick size to the specified value and
        returns the axis.

        Parameters
        ----------
        tick_size_inner : int
            Inner tick size

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> d3.axis_bottom().set_tick_size_inner(0)
        """
        self._tick_size_inner = tick_size_inner
        return self

    @property
    def tick_size_inner(self) -> int:
        return self._tick_size_inner

    def set_tick_size_outer(self, tick_size_outer: int) -> Axis:
        """
        Sets the outer tick size to the specified value and
        returns the axis.

        Parameters
        ----------
        tick_size_outer : int
            Outer tick size

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> d3.axis_bottom().set_tick_size_outer(0)
        """
        self._tick_size_outer = tick_size_outer
        return self

    @property
    def tick_size_outer(self) -> Axis:
        return self._tick_size_outer

    def set_tick_padding(self, tick_padding: int) -> Axis:
        """
        Sets the padding to the specified value in pixels.

        Parameters
        ----------
        tick_padding : int
            Tick padding in pixels

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> d3.axis_bottom().set_tick_padding(0)
        """
        self._tick_padding = tick_padding
        return self

    @property
    def tick_padding(self) -> Axis:
        return self._tick_padding

    def set_offset(self, offset: int | float) -> Axis:
        """
        Sets the pixel offset to the specified value in pixels.

        Parameters
        ----------
        offset : int | float
            Pixel offset

        Returns
        -------
        Axis
            Itself

        Examples
        --------

        >>> d3.axis_bottom().set_offset(0)
        """
        self._offset = offset
        return self

    @property
    def offset(self) -> int | float:
        return self._offset


def axis_top(scale: Type[Transformer | Sequential | Diverging]) -> Axis:
    """
    Builds a new top-oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.
    In this orientation, ticks are drawn above the horizontal domain path.

    Parameters
    ----------
    scale : Type[Transformer | Sequential | Diverging]
        Scaler

    Returns
    -------
    Axis
        Axis
    """
    return Axis(TOP, scale)


def axis_right(scale: Type[Transformer | Sequential | Diverging]) -> Axis:
    """
    Builds a new right-oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.
    In this orientation, ticks are drawn above the vertical domain path.

    Parameters
    ----------
    scale : Type[Transformer | Sequential | Diverging]
        Scaler

    Returns
    -------
    Axis
        Axis
    """
    return Axis(RIGHT, scale)


def axis_bottom(scale: Type[Transformer | Sequential | Diverging]) -> Axis:
    """
    Builds a new bottom-oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.
    In this orientation, ticks are drawn above the horizontal domain path.

    Parameters
    ----------
    scale : Type[Transformer | Sequential | Diverging]
        Scaler

    Returns
    -------
    Axis
        Axis
    """
    return Axis(BOTTOM, scale)


def axis_left(scale: Type[Transformer | Sequential | Diverging]) -> Axis:
    """
    Builds a new left-oriented axis generator for the given scale,
    with empty tick arguments, a tick size of 6 and padding of 3.
    In this orientation, ticks are drawn above the vertical domain path.

    Parameters
    ----------
    scale : Type[Transformer | Sequential | Diverging]
        Scaler

    Returns
    -------
    Axis
        Axis
    """
    return Axis(LEFT, scale)



from collections.abc import Callable
from datetime import datetime

from ..coloration.color import Color, color
from .constant import constant
from .date import interpolate_date
from .number import interpolate_number
from .number_array import interpolate_number_array, is_number_array
from .rgb import interpolate_rgb
from .string import interpolate_string


def interpolate_object(a, b) -> Callable:
    """
    Returns an interpolator between the two objects a and b.

    Returns
    -------
    Callable
        Interpolator
    """
    i = {}
    c = {}

    if a is None or not isinstance(a, dict):
        a = {}
    if b is None or not isinstance(b, dict):
        b = {}

    for k in b:
        if k in a:
            i[k] = interpolate(a.get(k), b.get(k))
        else:
            c[k] = b.get(k)

    def local_interpolate(t):
        for k in i:
            c[k] = i[k](t)
        return c

    return local_interpolate


def interpolate_array(a: list, b: list) -> Callable:
    """
    Returns an interpolator between the two arrays a and b.

    Parameters
    ----------
    a : list
        Array a
    b : list
        Array b

    Returns
    -------
    Callable
        Interpolator
    """
    return interpolate_number_array(a, b) if is_number_array(b) else generic_array(a, b)


def generic_array(a, b):
    nb = len(b) if b else 0
    na = min(nb, len(a)) if a else 0
    x = [interpolate(a[i], b[i]) for i in range(na)]
    c = list(b)

    def local_interpolate(t):
        for i in range(na):
            c[i] = x[i](t)
        return c

    return local_interpolate


def interpolate(a, b):
    """
    Returns an interpolator between the two arbitrary values a and b.
    """
    if b is None or isinstance(b, bool):
        return constant(b)
    if isinstance(b, (int, float)):
        return interpolate_number(a, b)
    if isinstance(b, str):
        c = color(b)
        if c:
            b = c
            return interpolate_rgb(a, b)
        else:
            return interpolate_string(a, b)
    if isinstance(b, Color):
        return interpolate_rgb(a, b)
    if isinstance(b, datetime):
        return interpolate_date(a, b)
    if is_number_array(b):
        return interpolate_number_array(a, b)
    if isinstance(b, (list, tuple)):
        return generic_array(a, b)
    if isinstance(b, dict):
        return interpolate_object(a, b)
    return interpolate_number(a, b)



from .array import (
    bin,
    extent,
    nice,
    tick_increment,
    tick_step,
    ticks,
)
from .axis import axis_bottom, axis_left, axis_right, axis_top
from .coloration import (
    color,
    cubehelix,
    gray,
    hcl,
    hsl,
    lab,
    lch,
    rgb,
)
from .format import (
    format_default_locale,
    format_prefix,
    precision_fixed,
    precision_prefix,
    precision_round,
)
from .format import (
    locale_format as format,
)
from .interpolate import (
    interpolate,
    interpolate_array,
    interpolate_cubehelix,
    interpolate_cubehelix_long,
    interpolate_date,
    interpolate_discrete,
    interpolate_hcl,
    interpolate_hcl_long,
    interpolate_hsl,
    interpolate_hsl_long,
    interpolate_hue,
    interpolate_lab,
    interpolate_number,
    interpolate_number_array,
    interpolate_object,
    interpolate_rgb,
    interpolate_rgb_basis,
    interpolate_rgb_basis_closed,
    interpolate_round,
    interpolate_string,
    interpolate_zoom,
)
from .path import Path as path
from .scale import (
    scale_band,
    scale_diverging,
    scale_diverging_log,
    scale_diverging_pow,
    scale_diverging_sqrt,
    scale_diverging_symlog,
    scale_identity,
    scale_linear,
    scale_log,
    scale_ordinal,
    scale_point,
    scale_pow,
    scale_quantile,
    scale_quantize,
    scale_radial,
    scale_sequential,
    scale_sequential_log,
    scale_sequential_pow,
    scale_sequential_quantile,
    scale_sequential_sqrt,
    scale_sequential_symlog,
    scale_sqrt,
    scale_symlog,
    scale_threshold,
    scale_time,
    tick_format,
)
from .selection import create
from .shape import Line as line
from .time import (
    time_day,
    time_friday,
    time_hour,
    time_millisecond,
    time_minute,
    time_monday,
    time_month,
    time_saturday,
    time_second,
    time_sunday,
    time_thursday,
    time_ticks,
    time_tuesday,
    time_wednesday,
    time_week,
    time_year,
)
from .time_format import (
    iso_format,
    iso_parse,
    time_format,
    time_format_locale,
    time_parse,
)

__all__ = [
    "axis_bottom",
    "axis_left",
    "axis_right",
    "axis_top",
    "bin",
    "color",
    "create",
    "cubehelix",
    "extent",
    "format",
    "format_default_locale",
    "format_prefix",
    "gray",
    "hcl",
    "hsl",
    "interpolate",
    "interpolate_array",
    "interpolate_cubehelix",
    "interpolate_cubehelix_long",
    "interpolate_date",
    "interpolate_discrete",
    "interpolate_hcl",
    "interpolate_hcl_long",
    "interpolate_hsl",
    "interpolate_hsl_long",
    "interpolate_hue",
    "interpolate_lab",
    "interpolate_number",
    "interpolate_number_array",
    "interpolate_object",
    "interpolate_rgb",
    "interpolate_rgb_basis",
    "interpolate_rgb_basis_closed",
    "interpolate_round",
    "interpolate_string",
    "interpolate_zoom",
    "iso_format",
    "iso_parse",
    "lab",
    "lch",
    "line",
    "nice",
    "path",
    "precision_fixed",
    "precision_prefix",
    "precision_round",
    "rgb",
    "scale_band",
    "scale_diverging",
    "scale_diverging_log",
    "scale_diverging_pow",
    "scale_diverging_sqrt",
    "scale_diverging_symlog",
    "scale_identity",
    "scale_linear",
    "scale_log",
    "scale_ordinal",
    "scale_point",
    "scale_pow",
    "scale_quantile",
    "scale_quantize",
    "scale_radial",
    "scale_sequential",
    "scale_sequential_log",
    "scale_sequential_pow",
    "scale_sequential_quantile",
    "scale_sequential_sqrt",
    "scale_sequential_symlog",
    "scale_sqrt",
    "scale_symlog",
    "scale_threshold",
    "scale_time",
    "tick_format",
    "tick_increment",
    "tick_step",
    "ticks",
    "time_day",
    "time_format",
    "time_format_locale",
    "time_friday",
    "time_hour",
    "time_millisecond",
    "time_minute",
    "time_monday",
    "time_month",
    "time_parse",
    "time_saturday",
    "time_second",
    "time_sunday",
    "time_thursday",
    "time_ticks",
    "time_tuesday",
    "time_wednesday",
    "time_week",
    "time_year",
]



import math
from collections.abc import Callable, Iterable
from inspect import signature
from typing import TypeVar

T = TypeVar("T")


# TODO : use min and max instead of doing manually
def extent(
    values: Iterable[T], accessor: Callable[[T, int, Iterable[T]], T] | None = None
) -> tuple[T, T]:
    """
    Returns the minimum and maximum value in
    the given iterable using natural order.

    Parameters
    ----------
    values : Iterable[T]
        Iterator
    accessor : Callable[[T, int, Iterable[T]], T] | None
        Accessor function

    Returns
    -------
    tuple[T, T]
        Minimum, maximum
    """
    mini = None
    maxi = None
    if accessor is None:
        for value in values:
            if value is not None and (isinstance(value, str) or not math.isnan(value)):
                if mini is None:
                    mini = maxi = value
                else:
                    if mini > value:
                        mini = value
                    if maxi < value:
                        maxi = value
    else:
        nargs = len(signature(accessor).parameters)
        for index, value in enumerate(values):
            args = [value, index, values][:nargs]
            new_value = accessor(*args)
            if new_value is not None and (
                isinstance(new_value, str) or not math.isnan(new_value)
            ):
                value = new_value
                if mini is None:
                    mini = maxi = value
                else:
                    if mini > value:
                        mini = value
                    if maxi < value:
                        maxi = value
    return [mini, maxi]  # TODO: remove brackets and fix tests



from __future__ import annotations

import math
from collections.abc import Callable
from datetime import datetime
from typing import Any, TypeVar, overload

from ..interpolate import interpolate, interpolate_round, piecewise
from .continuous import identity
from .init import init_interpolator
from .linear import LinearBase
from .log import LogBase, logp, powp, reflect, transform_log, transform_logn
from .pow import transform_pow, transform_sqrt
from .sequential import copy
from .symlog import transform_symlog

T = TypeVar("T")


class Diverging:
    """
    Diverging scales are similar to linear scales in that they map
    a continuous, numeric input domain to a continuous output range.
    Unlike linear scales, the input domain and output range of a diverging
    scale always have exactly three elements, and the output range is
    typically specified as an interpolator rather than an array of values.
    Diverging scales are typically used for a color encoding; see also
    d3-scale-chromatic. These scales do not expose invert and interpolate
    methods. There are also log, pow, and symlog variants of diverging scales.

    Parameters
    ----------
    t : Callable
        Transform function
    """

    def __init__(self, t: Callable):
        self.transform = t
        self._x0 = 0
        self._x1 = 0.5
        self._x2 = 1
        self._t0 = t(self._x0)
        self._t1 = t(self._x1)
        self._t2 = t(self._x2)
        self._k10 = 0 if self._t0 == self._t1 else 0.5 / (self._t1 - self._t0)
        self._k21 = 0 if self._t1 == self._t2 else 0.5 / (self._t2 - self._t1)
        self._s = -1 if self._t1 < self._t0 else 1
        self._interpolator = identity
        self._clamp = False
        self._unknown = None

    def __call__(self, x: int | float) -> T:
        """
        Given a value from the domain, returns the corresponding value from the range.

        Parameters
        ----------
        x : int | float
            Input value

        Returns
        -------
        T
            Corresponding value from the range
        """
        if isinstance(x, float) and math.isnan(x):
            return self._unknown
        else:
            x = self.transform(x)
            k = self._k10 if self._s * x < self._s * self._t1 else self._k21
            x = 0.5 + (x - self._t1) * k
            return self._interpolator(max(0, min(1, x)) if self._clamp else x)

    def set_domain(self, domain: list[int | float]) -> Diverging:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[int | float]
            Domain

        Returns
        -------
        Diverging
            Itself
        """
        self._x0, self._x1, self._x2 = map(float, list(domain)[:3])
        self._t0 = self.transform(self._x0)
        self._t1 = self.transform(self._x1)
        self._t2 = self.transform(self._x2)
        self._k10 = 0 if self._t0 == self._t1 else 0.5 / (self._t1 - self._t0)
        self._k21 = 0 if self._t1 == self._t2 else 0.5 / (self._t2 - self._t1)
        self._s = -1 if self._t1 < self._t0 else 1
        return self

    @property
    def domain(self) -> list[int | float]:
        return [self._x0, self._x1, self._x2]

    def set_clamp(self, clamp: bool) -> Diverging:
        """
        Enables or disables clamping accordingly.

        Parameters
        ----------
        clamp : bool
            Clamp value

        Returns
        -------
        Diverging
            Itself
        """
        self._clamp = bool(clamp)
        return self

    @property
    def clamp(self) -> bool:
        return self._clamp

    def set_interpolator(self, interpolator: Callable) -> Diverging:
        """
        Sets the scale’s interpolator to the specified function.

        Parameters
        ----------
        interpolator : Callable
            Interpolator function

        Returns
        -------
        Diverging
            Itself
        """
        self._interpolator = interpolator
        return self

    @property
    def interpolator(self) -> Callable:
        return self._interpolator

    def set_range(self, range_vals: list[T]) -> Diverging:
        """
        The given two-element array is converted
        to an interpolator function using interpolate

        Parameters
        ----------
        range_vals : list[T]
            Two values

        Returns
        -------
        Diverging
            Itself
        """
        self._r0 = float(range_vals[0])
        self._r1 = float(range_vals[1])
        self._r2 = float(range_vals[2])
        self._interpolator = piecewise(interpolate, [self._r0, self._r1, self._r2])
        return self

    def set_range_round(self, range_vals: list[T]) -> Diverging:
        """
        Sets the scale's range to the specified array of values
        and sets scale's interpolator to :code:`interpolate_round`.

        Parameters
        ----------
        range_vals : list[T]
            Range values

        Returns
        -------
        Diverging
            Itself
        """
        self._r0 = float(range_vals[0])
        self._r1 = float(range_vals[1])
        self._r2 = float(range_vals[2])
        self._interpolator = piecewise(
            interpolate_round, [self._r0, self._r1, self._r2]
        )
        return self

    @property
    def range(self) -> list[T]:
        return [self._interpolator(0), self._interpolator(0.5), self._interpolator(1)]

    def set_unknown(self, unknown: Any) -> Diverging:
        """
        Sets the output value of the scale for undefined
        or NaN input values.

        Parameters
        ----------
        unknown : Any
            Unknown value

        Returns
        -------
        Diverging
            Itself
        """
        self._unknown = unknown
        return self

    @property
    def unknown(self) -> Any:
        return self._unknown


class DivergingLinear(Diverging, LinearBase):
    def __init__(self):
        Diverging.__init__(self, identity)

    def copy(self):
        return copy(self, DivergingLinear())


class DivergingLog(Diverging, LogBase):
    def __init__(self):
        Diverging.__init__(self, identity)
        LogBase.__init__(self)
        self.transform = transform_log
        self.set_domain([0.1, 1, 10])

    def _rescale(self):
        self._logs = logp(self._base)
        self._pows = powp(self._base)
        d = self.domain[0]
        if isinstance(d, datetime):
            d = d.timestamp()
        if d < 0:
            self._logs = reflect(self._logs)
            self._pows = reflect(self._pows)
            self.transform = transform_logn
        else:
            self.transform = transform_log
        return self

    def set_domain(self, domain: list[int | float]) -> DivergingLog:
        """
        Sets the scale's domain to the specified array of numbers

        Parameters
        ----------
        domain : list[int | float]
            Domain

        Returns
        -------
        DivergingLog
            Itself
        """
        self._x0, self._x1, self._x2 = map(float, list(domain)[:3])
        self._rescale()
        super().set_domain(domain)
        return self

    def copy(self):
        return copy(self, DivergingLog()).base(self.base)


class DivergingSymlog(Diverging):
    def __init__(self, c: int | float = 1):
        self._c = c
        super().__init__(transform_symlog(self._c))

    def set_constant(self, c: int | float) -> DivergingSymlog:
        """
        Sets the symlog constant to the specified number and returns this scale.

        Parameters
        ----------
        c : int | float
            Constant value

        Returns
        -------
        DivergingSymlog
            Itself
        """
        self._c = float(c)
        self.transform = transform_symlog(self._c)
        self.rescale()
        return self

    def copy(self):
        return copy(self, DivergingSymlog()).set_constant(self.constant)


class DivergingPow(Diverging, LinearBase):
    def __init__(self, t: Callable = identity):
        super().__init__(t)
        self._exponent = 1

    def _rescale(self):
        if self._exponent == 1:
            self.transform = identity
            self.rescale()
            return self
        elif self._exponent == 0.5:
            self.transform = transform_sqrt
            self.rescale()
            return self
        else:
            self.transform = transform_pow(self._exponent)
            self.rescale()
            return self

    def set_exponent(self, exponent: int | float) -> DivergingPow:
        """
        Sets the scale's exponent value.

        Parameters
        ----------
        exponent : int | float
            Exponent value

        Returns
        -------
        DivergingPow
            Itself
        """
        self._exponent = float(exponent)
        return self._rescale()

    @property
    def exponent(self):
        return self._exponent

    def copy(self):
        return copy(self, DivergingPow()).set_exponent(self.exponent)


@overload
def scale_diverging() -> DivergingLinear: ...


@overload
def scale_diverging(interpolator: Callable) -> DivergingLinear: ...


@overload
def scale_diverging(
    domain: list[int | float], interpolator: Callable
) -> DivergingLinear: ...


def scale_diverging(*args):
    """
    Builds a new diverging scale with the specified
    domain and interpolator function or array.

    Parameters
    ----------
    domain : list[int | float]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingLinear
        Diverging object

    Examples
    --------

    >>> d3.scale_diverging([-1, 0, 1], d3.interpolate_RdBu)
    """
    scale = DivergingLinear()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_diverging_log() -> DivergingLinear: ...


@overload
def scale_diverging_log(interpolator: Callable) -> DivergingLinear: ...


@overload
def scale_diverging_log(
    domain: list[int | float], interpolator: Callable
) -> DivergingLinear: ...


def scale_diverging_log(*args):
    """
    Builds a new diverging scale with a logarithmic
    transform, analogous to a log scale.

    Parameters
    ----------
    domain : list[int | float]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingLog
        Diverging object

    Examples
    --------

    >>> d3.scale_diverging_log([1, 10, 100], d3.interpolate_RdBu)
    """
    scale = DivergingLog()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_diverging_symlog() -> DivergingLinear: ...


@overload
def scale_diverging_symlog(interpolator: Callable) -> DivergingLinear: ...


@overload
def scale_diverging_symlog(
    domain: list[int | float], interpolator: Callable
) -> DivergingLinear: ...


def scale_diverging_symlog(*args):
    """
    Builds a new diverging scale with a symmetric
    logarithmic transform, analogous to a symlog scale.

    Parameters
    ----------
    domain : list[int | float]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingSymlog
        Diverging object

    Examples
    --------

    >>> d3.scale_diverging_symlog([1, 10, 100], d3.interpolate_RdBu)
    """
    scale = DivergingSymlog()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_diverging_pow() -> DivergingLinear: ...


@overload
def scale_diverging_pow(interpolator: Callable) -> DivergingLinear: ...


@overload
def scale_diverging_pow(
    domain: list[int | float], interpolator: Callable
) -> DivergingLinear: ...


def scale_diverging_pow(*args):
    """
    Builds a new diverging scale with an exponential
    transform, analogous to a power scale.


    Parameters
    ----------
    domain : list[int | float]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingPow
        Diverging object

    Examples
    --------

    >>> d3.scale_diverging_pow([0, 1, 10], d3.interpolate_RdBu)
    """
    scale = DivergingPow()
    if len(args) == 1:
        return init_interpolator(scale, interpolator=args[0])
    elif len(args) == 2:
        domain, interpolator = args
        return init_interpolator(scale, domain=domain, interpolator=interpolator)
    return init_interpolator(scale)


@overload
def scale_diverging_sqrt() -> DivergingLinear: ...


@overload
def scale_diverging_sqrt(interpolator: Callable) -> DivergingLinear: ...


@overload
def scale_diverging_sqrt(
    domain: list[int | float], interpolator: Callable
) -> DivergingLinear: ...


def scale_diverging_sqrt(*args):
    """
    Builds a new diverging scale with a square-root
    transform, analogous to a sqrt scale.

    Parameters
    ----------
    domain : list[int | float]
        Array of numbers
    interpolator : Callable
        Interpolator function

    Returns
    -------
    DivergingSqrt
        Diverging object

    Examples
    --------

    >>> d3.scale_diverging_sqrt([0, 1, 10], d3.interpolate_RdBu)
    """
    return scale_diverging_pow(*args).exponent(0.5)



from ..interpolate import interpolate_rgb_basis_closed
from ..scale import scale_sequential
from .colors import color


def ramp_closed(rang):
    return scale_sequential(interpolate_rgb_basis_closed(colors(rang))).clamp(True)



from pathlib import Path

ROOT = Path()
FILE = Path("scale.py")

lines = FILE.read_text().split("\n")
current_file = None
content = []
has_changed = False
for line in lines:
    if "# src/" in line:
        if file := current_file:
            file.parent.mkdir(parents=True, exist_ok=True)
            file.write_text("\n".join(content))
        current_file = ROOT / "/".join(line.split("/")[1:])
        content = []
        continue
    content.append(line)

if (file := current_file) and content:
    file.parent.mkdir(parents=True, exist_ok=True)
    file.write_text("\n".join(content))



from collections.abc import Callable
from typing import TypeVar

T = TypeVar("T")


def interpolate_number(a: T, b: T) -> Callable[[T], T]:
    """
    Returns an interpolator between the two numbers a and b.

    Parameters
    ----------
    a : T
        a value
    b : T
        b value

    Returns
    -------
    Callable[[T], T]
        Interpolator
    """
    a, b = float(a), float(b)
    return lambda t: a * (1 - t) + b * t



from __future__ import annotations

import math
from bisect import bisect
from collections.abc import Callable, Iterable
from inspect import signature
from typing import Any

from .extent import extent
from .nice import nice
from .threshold import threshold_sturges
from .ticks import tick_increment, ticks


def identity(x, *args):
    return x


def constant(*obj):
    if len(obj) == 1:
        obj = obj[0]

    def wrapper(*args):
        return obj

    return wrapper


class Bin:
    """
    Bin quantitative values into consecutive, non-overlapping
    intervals, as in histograms
    """

    def __init__(self):
        self._list = []
        self.x0 = None
        self.x1 = None

    def __len__(self):
        return len(self._list)

    def __getitem__(self, key):
        return self._list[key]

    def __setitem__(self, key, item):
        self._list[key] = item

    def append(self, item):
        self._list.append(item)

    def __str__(self):
        return f"Bin({self._list}, x0={self.x0}, x1={self.x1})"

    def __repr__(self):
        return str(self)

    def __eq__(self, bin):
        return self._list == bin._list and self.x0 == self.x0 and self.x1 == self.x1


class bin:
    """
    Bin generator with the default settings
    """

    def __init__(self):
        self._value = identity
        self._domain = extent
        self._threshold = threshold_sturges

    def __call__(self, data: Iterable[int | float]) -> list[Bin]:
        """
        Returns an array of bins, where each bin is an array
        containing the associated elements from the input data.

        Parameters
        ----------
        data : Iterable[int | float]
            Data samples

        Returns
        -------
        list[Bin]
            Array of bins
        """
        if not isinstance(data, list):
            data = list(data)

        n = len(data)
        step = math.nan
        values = [None] * n

        nargs = len(signature(self._value).parameters)
        for i in range(n):
            args = [data[i], i, data][:nargs]
            values[i] = self._value(*args)

        xz = self._domain(values)
        x0, x1 = xz[0], xz[1]
        tz = self._threshold(values, x0, x1)

        if not isinstance(tz, (list, tuple)):
            max_ = x1
            tn = int(tz)
            if self._domain == extent:
                x0, x1 = nice(x0, x1, tn)
            tz = ticks(x0, x1, tn)

            if tz[0] <= x0:
                step = tick_increment(x0, x1, tn)

            if tz[-1] >= x1:
                if max_ >= x1 and self._domain == extent:
                    step = tick_increment(x0, x1, tn)
                    if math.isfinite(step):
                        if step > 0:
                            x1 = (math.floor(x1 / step) + 1) * step
                        elif step < 0:
                            x1 = (math.ceil(x1 * -step) + 1) / -step
                else:
                    tz.pop()

        m = len(tz)
        a = 0
        b = m
        while a < len(tz) and tz[a] <= x0:
            a += 1
        while b >= 0 and tz[b - 1] > x1:
            b -= 1
        if a or b < m:
            tz = tz[a:b]
            m = b - a

        bins = [None] * (m + 1)

        for i in range(m + 1):
            bins[i] = Bin()
            bins[i].x0 = tz[i - 1] if i > 0 else x0
            bins[i].x1 = tz[i] if i < m else x1

        if math.isfinite(step):
            if step > 0:
                for i in range(n):
                    if (x := values[i]) is not None and x0 <= x <= x1:
                        bins[min(m, math.floor((x - x0) / step))].append(data[i])
            elif step < 0:
                for i in range(n):
                    if (x := values[i]) is not None and x0 <= x <= x1:
                        j = math.floor((x0 - x) * step)
                        if j < len(tz):
                            bins[min(m, j + (tz[j] <= x))].append(data[i])
                        else:
                            bins[m].append(data[i])
        else:
            for i in range(n):
                if (x := values[i]) is not None and x0 <= x <= x1:
                    bins[bisect(tz, x, 0, m)].append(data[i])

        return bins

    def value(self, obj: Callable | Any | None = None) -> Callable | bin:
        """
        Set value to the given object or return the current value

        Parameters
        ----------
        obj : Callable | Any | None
            Object or function

        Returns
        -------
        Callable | bin
            Current value or updated self
        """
        if obj is None:
            return self._value
        elif callable(obj):
            self._value = obj
            return self
        else:
            self._value = constant(obj)
            return self

    def domain(self, obj: Callable | Any | None = None) -> Callable | bin:
        """
        Set domain to the given object or return the current domain

        Parameters
        ----------
        obj : Callable | Any | None
            Object or function

        Returns
        -------
        Callable | bin
            Current domain or updated self
        """
        if obj is None:
            return self._domain
        elif callable(obj):
            self._domain = obj
            return self
        else:
            self._domain = constant(obj[0], obj[1])
            return self

    def thresholds(self, obj: Callable | Any | None = None) -> Callable | bin:
        """
        Set thresholds to the given object or return the current thresholds

        Parameters
        ----------
        obj : Callable | Any | None
            Object or function

        Returns
        -------
        Callable | bin
            Current thresholds or updated self
        """
        if obj is None:
            return self._threshold
        elif callable(obj):
            self._threshold = obj
            return self
        else:
            self._threshold = constant(obj)
            return self



from .basis import basis


def basis_closed(values):
    n = len(values)

    def interpolate(t):
        i = int((t % 1) * n)
        v0 = values[(i + n - 1) % n]
        v1 = values[i % n]
        v2 = values[(i + 1) % n]
        v3 = values[(i + 2) % n]
        return basis((t - i / n) * n, v0, v1, v2, v3)

    return interpolate



from .bin import bin
from .extent import extent
from .nice import nice
from .ticks import tick_increment, tick_step, ticks

__all__ = [
    "bin",
    "extent",
    "nice",
    "tick_increment",
    "tick_step",
    "ticks",
]



from inspect import signature


def attr_function(name, value):
    nargs = len(signature(value).parameters)

    def callback(node, data, i, group):
        args = [data, i, group][:nargs]
        node.set(name, str(value(*args)))

    return callback


def attr_constant(name, value):
    def callback(node, data, i, group):
        node.set(name, str(value))

    return callback



# src/hue.py
from .color import hue


def hue_interpolator(a, b):
    i = hue(float(a), float(b))

    def interpolate(t):
        x = i(t)
        return x - 360 * math.floor(x / 360)

    return interpolate


# src/rgb.py
from d3_color import rgb as colorRgb

from .basis import basis
from .basisClosed import basisClosed
from .color import gamma, nogamma


def rgbGamma(y):
    color = gamma(y)

    def rgb(start, end):
        start = colorRgb(start)
        end = colorRgb(end)
        r = color(start.r, end.r)
        g = color(start.g, end.g)
        b = color(start.b, end.b)
        opacity = nogamma(start.opacity, end.opacity)

        def interpolate(t):
            start.r = r(t)
            start.g = g(t)
            start.b = b(t)
            start.opacity = opacity(t)
            return str(start)

        return interpolate

    rgb.gamma = rgbGamma
    return rgb


rgb = rgbGamma(1)


def rgbSpline(spline):
    def interpolate(colors):
        n = len(colors)
        r = [0] * n
        g = [0] * n
        b = [0] * n

        for i, color in enumerate(colors):
            color = colorRgb(color)
            r[i] = color.r or 0
            g[i] = color.g or 0
            b[i] = color.b or 0

        r_spline = spline(r)
        g_spline = spline(g)
        b_spline = spline(b)

        color = colorRgb()
        color.opacity = 1

        def interpolate(t):
            color.r = r_spline(t)
            color.g = g_spline(t)
            color.b = b_spline(t)
            return str(color)

        return interpolate

    return interpolate


rgbBasis = rgbSpline(basis)
rgbBasisClosed = rgbSpline(basisClosed)

# src/object.py
from .value import value


def object_interpolator(a, b):
    i = {}
    c = {}

    if a is None or not isinstance(a, dict):
        a = {}
    if b is None or not isinstance(b, dict):
        b = {}

    for k in b:
        if k in a:
            i[k] = value(a[k], b[k])
        else:
            c[k] = b[k]

    def interpolate(t):
        for k in i:
            c[k] = i[k](t)
        return c

    return interpolate


# src/numberArray.py
import numpy as np


def numberArray(a, b):
    if b is None:
        b = []
    n = min(len(b), len(a)) if a else 0
    c = b.copy()

    def interpolate(t):
        for i in range(n):
            c[i] = a[i] * (1 - t) + b[i] * t
        return c

    return interpolate


def isNumberArray(x):
    return isinstance(x, (np.ndarray, list)) and all(
        isinstance(i, (int, float)) for i in x
    )


# src/hcl.py
from d3_color import hcl as colorHcl

from .color import color, hue


def hcl(hue_func):
    def interpolator(start, end):
        start = colorHcl(start)
        end = colorHcl(end)
        h = hue_func(start.h, end.h)
        c = color(start.c, end.c)
        l = color(start.l, end.l)
        opacity = color(start.opacity, end.opacity)

        def interpolate(t):
            start.h = h(t)
            start.c = c(t)
            start.l = l(t)
            start.opacity = opacity(t)
            return str(start)

        return interpolate

    return interpolator


hcl_default = hcl(hue)
hclLong = hcl(color)


# src/number.py
def number(a, b):
    a, b = float(a), float(b)
    return lambda t: a * (1 - t) + b * t


# src/cubehelix.py
from d3_color import cubehelix as colorCubehelix

from .color import color, hue


def cubehelix(hue_func):
    def cubehelixGamma(y):
        y = float(y)

        def interpolator(start, end):
            start = colorCubehelix(start)
            end = colorCubehelix(end)
            h = hue_func(start.h, end.h)
            s = color(start.s, end.s)
            l = color(start.l, end.l)
            opacity = color(start.opacity, end.opacity)

            def interpolate(t):
                start.h = h(t)
                start.s = s(t)
                start.l = l(t**y)
                start.opacity = opacity(t)
                return str(start)

            return interpolate

        interpolator.gamma = cubehelixGamma
        return interpolator

    return cubehelixGamma(1)


cubehelix_default = cubehelix(hue)
cubehelixLong = cubehelix(color)

# src/piecewise.py
from .value import value as default_value


def piecewise(interpolate, values=None):
    if values is None:
        values = interpolate
        interpolate = default_value

    i = 0
    n = len(values) - 1
    v = values[0]
    I = [interpolate(v, values[i + 1]) for i in range(n)]

    def interpolator(t):
        i = max(0, min(n - 1, int(t * n)))
        return I[i](t * n - i)

    return interpolator


# src/round.py
def round_interpolator(a, b):
    a, b = float(a), float(b)
    return lambda t: round(a * (1 - t) + b * t)


# src/value.py
from d3_color import color

from .array import genericArray
from .constant import constant
from .date import date
from .number import number
from .numberArray import isNumberArray, numberArray
from .object import object_interpolator
from .rgb import rgb
from .string import string


def value(a, b):
    if b is None or isinstance(b, bool):
        return constant(b)
    if isinstance(b, (int, float)):
        return number
    if isinstance(b, str):
        c = color(b)
        return rgb if c else string
    if isinstance(b, color):
        return rgb
    if isinstance(b, datetime.date):
        return date
    if isNumberArray(b):
        return numberArray
    if isinstance(b, (list, tuple)):
        return genericArray
    if not hasattr(b, "valueOf") and not hasattr(b, "__str__") or math.isnan(b):
        return object_interpolator
    return number


# src/hsl.py
from d3_color import hsl as colorHsl

from .color import color, hue


def hsl(hue_func):
    def interpolator(start, end):
        start = colorHsl(start)
        end = colorHsl(end)
        h = hue_func(start.h, end.h)
        s = color(start.s, end.s)
        l = color(start.l, end.l)
        opacity = color(start.opacity, end.opacity)

        def interpolate(t):
            start.h = h(t)
            start.s = s(t)
            start.l = l(t)
            start.opacity = opacity(t)
            return str(start)

        return interpolate

    return interpolator


hsl_default = hsl(hue)
hslLong = hsl(color)


# src/discrete.py
def discrete(range_):
    n = len(range_)
    return lambda t: range_[max(0, min(n - 1, int(t * n)))]


# src/string.py
import re

from .number import number

reA = re.compile(r"[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?")
reB = re.compile(reA.pattern)


def zero(b):
    return lambda: b


def one(b):
    return lambda t: str(b(t))


def string(a, b):
    a, b = str(a), str(b)

    bi = 0
    s = []
    q = []

    for am in reA.finditer(a):
        bm = reB.search(b, bi)
        if bm:
            if bm.start() > bi:
                s.append(b[bi : bm.start()])

            if am.group() == bm.group():
                s.append(bm.group())
            else:
                s.append(None)
                q.append(
                    {"i": len(s) - 1, "x": number(float(am.group()), float(bm.group()))}
                )

            bi = bm.end()

    if bi < len(b):
        s.append(b[bi:])

    if len(s) < 2:
        return one(q[0]["x"]) if q else zero(b)

    def interpolator(t):
        for o in q:
            s[o["i"]] = str(o["x"](t))
        return "".join(s)

    return interpolator


# src/constant.py
def constant(x):
    return lambda: x


# src/array.py
from .numberArray import isNumberArray, numberArray
from .value import value


def array(a, b):
    return numberArray(a, b) if isNumberArray(b) else genericArray(a, b)


def genericArray(a, b):
    nb = len(b) if b else 0
    na = min(nb, len(a)) if a else 0
    x = [value(a[i], b[i]) for i in range(na)]
    c = list(b)

    def interpolate(t):
        for i in range(na):
            c[i] = x[i](t)
        return c

    return interpolate


# src/basisClosed.py
from .basis import basis


def basisClosed(values):
    n = len(values)

    def interpolate(t):
        i = int((t % 1) * n)
        v0 = values[(i + n - 1) % n]
        v1 = values[i % n]
        v2 = values[(i + 1) % n]
        v3 = values[(i + 2) % n]
        return basis((t - i / n) * n, v0, v1, v2, v3)

    return interpolate


# src/zoom.py
import math

epsilon2 = 1e-12


def cosh(x):
    return (math.exp(x) + 1 / math.exp(x)) / 2


def sinh(x):
    return (math.exp(x) - 1 / math.exp(x)) / 2


def tanh(x):
    return (math.exp(2 * x) - 1) / (math.exp(2 * x) + 1)


def zoomRho(rho, rho2, rho4):
    def zoom(p0, p1):
        ux0, uy0, w0 = p0
        ux1, uy1, w1 = p1
        dx = ux1 - ux0
        dy = uy1 - uy0
        d2 = dx * dx + dy * dy

        if d2 < epsilon2:
            S = math.log(w1 / w0) / rho

            def interpolator(t):
                return [ux0 + t * dx, uy0 + t * dy, w0 * math.exp(rho * t * S)]
        else:
            d1 = math.sqrt(d2)
            b0 = (w1 * w1 - w0 * w0 + rho4 * d2) / (2 * w0 * rho2 * d1)
            b1 = (w1 * w1 - w0 * w0 - rho4 * d2) / (2 * w1 * rho2 * d1)
            r0 = math.log(math.sqrt(b0 * b0 + 1) - b0)
            r1 = math.log(math.sqrt(b1 * b1 + 1) - b1)
            S = (r1 - r0) / rho

            def interpolator(t):
                s = t * S
                coshr0 = cosh(r0)
                u = w0 / (rho2 * d1) * (coshr0 * tanh(rho * s + r0) - sinh(r0))
                return [ux0 + u * dx, uy0 + u * dy, w0 * coshr0 / cosh(rho * s + r0)]

        interpolator.duration = S * 1000 * rho / math.sqrt(2)
        return interpolator

    def set_rho(new_rho):
        new_rho = max(1e-3, float(new_rho))
        return zoomRho(new_rho, new_rho * new_rho, new_rho**4)

    zoom.rho = set_rho
    return zoom


zoom = zoomRho(math.sqrt(2), 2, 4)


# src/basis.py
def basis(t1, v0, v1, v2, v3):
    t2 = t1 * t1
    t3 = t2 * t1
    return (
        (1 - 3 * t1 + 3 * t2 - t3) * v0
        + (4 - 6 * t2 + 3 * t3) * v1
        + (1 + 3 * t1 + 3 * t2 - 3 * t3) * v2
        + t3 * v3
    ) / 6


def basis_interpolator(values):
    n = len(values) - 1

    def interpolate(t):
        i = 0 if t <= 0 else (n - 1 if t >= 1 else int(t * n))
        v1 = values[i]
        v2 = values[i + 1]
        v0 = values[i - 1] if i > 0 else 2 * v1 - v2
        v3 = values[i + 2] if i < n - 1 else 2 * v2 - v1
        return basis((t - i / n) * n, v0, v1, v2, v3)

    return interpolate


# src/color.py
import math

from .constant import constant


def linear(a, d):
    return lambda t: a + t * d


def exponential(a, b, y):
    return lambda t: math.pow(a + t * (math.pow(b, y) - a), 1 / y)


def hue(a, b):
    d = b - a
    return (
        linear(a, d if abs(d) <= 180 else d - 360 * round(d / 360))
        if d
        else constant(math.isnan(a) and b or a)
    )


def gamma(y):
    y = float(y)
    if y == 1:
        return nogamma
    else:
        return (
            lambda a, b: exponential(a, b, y)
            if b - a
            else constant(math.isnan(a) and b or a)
        )


def nogamma(a, b):
    d = b - a
    return linear(a, d) if d else constant(math.isnan(a) and b or a)


# src/lab.py
from d3_color import lab as colorLab

from .color import color


def lab(start, end):
    start = colorLab(start)
    end = colorLab(end)
    l = color(start.l, end.l)
    a = color(start.a, end.a)
    b = color(start.b, end.b)
    opacity = color(start.opacity, end.opacity)

    def interpolate(t):
        start.l = l(t)
        start.a = a(t)
        start.b = b(t)
        start.opacity = opacity(t)
        return str(start)

    return interpolate


# src/date.py
import datetime


def date_interpolator(a, b):
    d = datetime.datetime.now()
    a, b = a.timestamp(), b.timestamp()

    def interpolate(t):
        d = datetime.datetime.fromtimestamp(a * (1 - t) + b * t)
        return d

    return interpolate


# src/quantize.py
def quantize(interpolator, n):
    samples = [interpolator(i / (n - 1)) for i in range(n)]
    return samples


# src/transform/parse.py
import math

from .decompose import decompose, identity


def parseCss(value):
    try:
        from cssselect import parser

        m = parser.parse_style_attribute(value)[0]
        return decompose(m.a, m.b, m.c, m.d, m.e, m.f)
    except ImportError:
        print("cssselect library not found. Install it for CSS parsing support.")
        return identity


def parseSvg(value):
    if value is None:
        return identity
    try:
        from svg.path import parse_path

        path = parse_path(value)
        m = path.transform()
        return decompose(m.a, m.b, m.c, m.d, m.e, m.f)
    except ImportError:
        print("svg.path library not found. Install it for SVG parsing support.")
        return identity


# src/transform/decompose.py
import math

degrees = 180 / math.pi

identity = {
    "translateX": 0,
    "translateY": 0,
    "rotate": 0,
    "skewX": 0,
    "scaleX": 1,
    "scaleY": 1,
}


def decompose(a, b, c, d, e, f):
    scaleX = math.sqrt(a * a + b * b)
    scaleY = math.sqrt(c * c + d * d)

    if scaleX:
        a /= scaleX
        b /= scaleX

    skewX = a * c + b * d
    if skewX:
        c -= a * skewX
        d -= b * skewX

    if scaleY:
        c /= scaleY
        d /= scaleY
        skewX /= scaleY

    if a * d < b * c:
        a, b, c, d = -a, -b, -c, -d
        scaleX, scaleY = -scaleX, -scaleY

    return {
        "translateX": e,
        "translateY": f,
        "rotate": math.atan2(b, a) * degrees,
        "skewX": math.atan(skewX) * degrees,
        "scaleX": scaleX,
        "scaleY": scaleY,
    }


# src/transform/index.py
from ..number import number
from .parse import parseCss, parseSvg


def interpolateTransform(parse, pxComma, pxParen, degParen):
    def pop(s):
        return s.pop() + " " if s else ""

    def translate(xa, ya, xb, yb, s, q):
        if xa != xb or ya != yb:
            s.extend(["translate(", None, pxComma, None, pxParen])
            q.extend(
                [
                    {"i": len(s) - 4, "x": number(xa, xb)},
                    {"i": len(s) - 2, "x": number(ya, yb)},
                ]
            )
        elif xb or yb:
            s.append(f"translate({xb}{pxComma}{yb}{pxParen}")

    def rotate(a, b, s, q):
        if a != b:
            if a - b > 180:
                b += 360
            elif b - a > 180:
                a += 360
            q.append({"i": len(s), "x": number(a, b)})
            s.extend([pop(s) + "rotate(", None, degParen])
        elif b:
            s.append(f"{pop(s)}rotate({b}{degParen}")

    def skewX(a, b, s, q):
        if a != b:
            q.append({"i": len(s), "x": number(a, b)})
            s.extend([pop(s) + "skewX(", None, degParen])
        elif b:
            s.append(f"{pop(s)}skewX({b}{degParen}")

    def scale(xa, ya, xb, yb, s, q):
        if xa != xb or ya != yb:
            s.extend([pop(s) + "scale(", None, ",", None, ")"])
            q.extend(
                [
                    {"i": len(s) - 4, "x": number(xa, xb)},
                    {"i": len(s) - 2, "x": number(ya, yb)},
                ]
            )
        elif xb != 1 or yb != 1:
            s.append(f"{pop(s)}scale({xb},{yb})")

    def interpolator(a, b):
        s = []
        q = []
        a = parse(a)
        b = parse(b)
        translate(
            a["translateX"], a["translateY"], b["translateX"], b["translateY"], s, q
        )
        rotate(a["rotate"], b["rotate"], s, q)
        skewX(a["skewX"], b["skewX"], s, q)
        scale(a["scaleX"], a["scaleY"], b["scaleX"], b["scaleY"], s, q)

        def interpolate(t):
            for o in q:
                s[o["i"]] = o["x"](t)
            return "".join(s)

        return interpolate

    return interpolator


interpolateTransformCss = interpolateTransform(parseCss, "px, ", "px)", "deg)")
interpolateTransformSvg = interpolateTransform(parseSvg, ", ", ")", ")")



from __future__ import annotations

import math
from typing import TypeVar, overload

from .continuous import Transformer, copy
from .init import init_range
from .linear import LinearBase

T = TypeVar("T")


def sign(x):
    return -1 if x < 0 else 1


def transform_symlog(c):
    return lambda x: sign(x) * math.log1p(abs(x / c))


def transform_symexp(c):
    return lambda x: sign(x) * math.expm1(abs(x)) * c


class ScaleSymlog(Transformer, LinearBase):
    def __init__(self, c=1):
        self._c = c
        super().__init__(transform_symlog(self._c), transform_symexp(self._c))

    def set_constant(self, c: int | float) -> ScaleSymlog:
        """
        Sets the symlog constant to the specified number and returns this scale.

        Parameters
        ----------
        c : int | float
            Constant value

        Returns
        -------
        ScaleSymlog
            Itself
        """
        self._c = float(c)
        self.transform = transform_symlog(self._c)
        self.untransform = transform_symexp(self._c)
        self.rescale()
        return self

    @property
    def constant(self):
        return self._c

    def copy(self):
        return copy(self, ScaleSymlog()).set_constant(self.constant)


@overload
def scale_symlog() -> ScaleSymlog: ...


@overload
def scale_symlog(range_vals: list[T]) -> ScaleSymlog: ...


@overload
def scale_symlog(domain: list[int | float], range_vals: list[T]) -> ScaleSymlog: ...


def scale_symlog(*args):
    """
    Builds a new continuous scale with the specified domain and
    range, the constant 1, the default interpolator and clamping
    disabled.

    Parameters
    ----------
    domain : list[int | float]
        Array of numbers
    range_vals : list[int | float]
        Array of values

    Returns
    -------
    ScaleSymlog
        Scale object

    Examples
    --------

    >>> d3.scale_symlog([0, 100], [0, 960])
    """
    scale = ScaleSymlog()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



def init_range(obj, domain=None, range_vals=None):
    if domain is None and range_vals is None:
        return obj
    elif domain is None:
        return obj.set_range(range_vals)
    else:
        return obj.set_range(range_vals).set_domain(domain)


def init_interpolator(obj, domain=None, interpolator=None):
    if domain is None and interpolator is None:
        return obj
    elif domain is not None:
        obj.set_domain(domain)
        if callable(domain):
            obj.set_interpolator(domain)
        else:
            obj.set_range(domain)
    else:
        if callable(interpolator):
            obj.set_interpolator(interpolator)
        else:
            obj.set_range(interpolator)
    return obj



def basis(t1, v0, v1, v2, v3):
    t2 = t1 * t1
    t3 = t2 * t1
    return (
        (1 - 3 * t1 + 3 * t2 - t3) * v0
        + (4 - 6 * t2 + 3 * t3) * v1
        + (1 + 3 * t1 + 3 * t2 - 3 * t3) * v2
        + t3 * v3
    ) / 6


def basis_interpolator(values):
    n = len(values) - 1

    def interpolate(t):
        i = 0 if t <= 0 else (n - 1 if t >= 1 else int(t * n))
        v1 = values[i]
        v2 = values[i + 1]
        v0 = values[i - 1] if i > 0 else 2 * v1 - v2
        v3 = values[i + 2] if i < n - 1 else 2 * v2 - v1
        return basis((t - i / n) * n, v0, v1, v2, v3)

    return interpolate



import math
from statistics import stdev


def threshold_scott(values, mini, maxi):
    values = [v for v in values if v is not None]
    c = len(values)
    if len(values) <= 1:
        d = 0
    else:
        d = stdev(values)
    return math.ceil((maxi - mini) * (c ** (1 / 3)) / (3.49 * d)) if c and d else 1



import math
from statistics import quantiles


def threshold_freedman_diaconis(values, mini, maxi):
    values = [v for v in values if v is not None]
    c = len(values)
    if len(values) <= 1:
        d = 0
    else:
        q = quantiles(values, n=4, method="inclusive")
        d = q[2] - q[0]
    return math.ceil((maxi - mini) / (2 * d * (c ** (-1 / 3)))) if c and d else 1



from __future__ import annotations

from collections.abc import Callable, Iterable
from inspect import signature

from ..selection.selection import Selection
from .constant import constant
from .curves.linear import LinearCurve
from .path import WithPath
from .point import x as point_x
from .point import y as point_y


class Line(WithPath):
    """
    The line generator produces a spline or polyline as in a line chart.
    Lines also appear in many other visualization types, such as the links
    in hierarchical edge bundling.

    Parameters
    ----------
    x : Callable | None
        x accessor function for data points
    y : Callable | None
        y accessor function for data points

    Returns
    -------
    Line
        New line generator
    """

    def __init__(self, x: Callable | None = None, y: Callable | None = None):
        super().__init__()
        self._defined = constant(True)
        self._context = None
        self._curve = LinearCurve
        self._output = None

        if x is None:
            self._x = point_x
        elif callable(x):
            self._x = x
        else:
            self._x = constant(x)

        if y is None:
            self._y = point_y
        elif callable(y):
            self._y = y
        else:
            self._y = constant(y)

    def __call__(self, data: Iterable) -> str | None:
        """
        Generate a line for the given list of data

        Parameters
        ----------
        data : Iterable
            Data values

        Returns
        -------
        str | None
            Generated line if the line is not associated to a context

        Examples
        --------

        >>> svg.append("path").attr("d", line(data)).attr("stroke", "currentColor")
        """
        data = list(data)
        n = len(data)
        defined0 = False

        if self._context is None:
            buffer = self._path()
            self._output = self._curve(buffer)

        xnargs = len(signature(self._x).parameters)
        ynargs = len(signature(self._y).parameters)
        for i in range(n):
            d = data[i]
            if i == n or self._defined(d, i, data) != defined0:
                defined0 = not defined0
                if defined0:
                    self._output.line_start()
                else:
                    self._output.line_end()
            if defined0:
                xargs = [d, i, data][:xnargs]
                yargs = [d, i, data][:ynargs]
                self._output.point(self._x(*xargs), self._y(*yargs))

        i += 1
        defined0 = not defined0
        if defined0:
            self._output.line_start()
        else:
            self._output.line_end()
        if defined0:
            xargs = [d, i, data][:xnargs]
            yargs = [d, i, data][:ynargs]
            self._output.point(self._x(xargs), self._y(yargs))

        if buffer:
            self._output = None
            return str(buffer) or None

    def x(self, x: Callable | int | float | None) -> Line:
        """
        Sets x accessor function

        Parameters
        ----------
        x : Callable | int | float | None
            x accessor function

        Returns
        -------
        Line
            Itself
        """
        if x is None:
            self._x = point_x
        elif callable(x):
            self._x = x
        else:
            self._x = constant(x)
        return self

    @property
    def fx(self):
        return self._x

    def y(self, y: Callable | int | float | None) -> Line:
        """
        Sets y accessor function

        Parameters
        ----------
        y : Callable | int | float | None
            y accessor function

        Returns
        -------
        Line
            Itself
        """
        if y is None:
            self._y = point_y
        elif callable(y):
            self._y = y
        else:
            self._y = constant(y)
        return self

    @property
    def fy(self):
        return self._y

    def defined(self, defined: Callable | int | float | None) -> Line:
        """
        Sets defined accessor

        When a line is generated, the defined accessor will be invoked
        for each element in the input data array, being passed the element
        :code:`d`, the index :code:`i`, and the array :code:`data` as three
        arguments. If the given element is defined (i.e., if the defined accessor
        returns a truthy value for this element), the x and y accessors will
        subsequently be evaluated and the point will be added to the current
        line segment. Otherwise, the element will be skipped, the current line
        segment will be ended, and a new line segment will be generated for the
        next defined point.

        Parameters
        ----------
        defined : Callable | int | float | None
            defined accessor function

        Returns
        -------
        Line
            Itself
        """
        if defined is None:
            self._defined = defined
        elif callable(defined):
            self._defined = defined
        else:
            self._defined = constant(bool(defined))
        return self

    @property
    def accessor_defined(self):
        return self._defined

    def curve(self, curve: Callable | None) -> Line:
        """
        Sets curve factory.

        Parameters
        ----------
        curve : Callable | None
            Curve factory function

        Returns
        -------
        Line
            Itself
        """
        self._curve = curve
        if self._context is not None:
            self._output = self._curve(self._context)
        return self

    @property
    def fcurve(self):
        return self._curve

    def context(self, context: Selection | None) -> Line:
        """
        Sets the context.

        Parameters
        ----------
        context : Selection | None
            Selection

        Returns
        -------
        Line
            Itself
        """
        if context is None:
            self._context = None
            self._output = None
        else:
            self._context = context
            self._output = self._curve(self._context)
        return self

    @property
    def own_context(self):
        return self._context
